@@ -29,6 +29,8 @@ from rich.syntax import Syntax
 # internal modules
 from hledger_utils.utils import (
     find_hledger_command,
+    handle_keyboardinterrupt,
+    handle_exception,
     sanitize_filename,
     splitlines,
     two_at_a_time,
@@ -117,6 +119,12 @@ hledger_parser.add_argument("-O", "--output-format")
 hledger_parser.add_argument("-o", "--output-file")
 
 
+@handle_keyboardinterrupt(
+    console=console,
+    exitcode=handle_exception(
+        0, int, os.environ.get("HLEDGER_EDIT_KEYBOARDINTERRUPT_EXITCODE")
+    ),
+)
 def cli(cli_args=sys.argv[1:]):
     args, hledger_args = parser.parse_known_args(cli_args)
     logging.basicConfig(
@@ -163,7 +171,7 @@ def cli(cli_args=sys.argv[1:]):
         sys.exit(0)
 
     def file_start_end():
-        for i, tx in enumerate(transactions):
+        for i, tx in enumerate(transactions, start=1):
             txshort = {
                 k: v
                 for k, v in tx.items()
@@ -171,7 +179,7 @@ def cli(cli_args=sys.argv[1:]):
             }
             if not (tsourcepos := tx.get("tsourcepos")):
                 logger.info(
-                    f"Transaction result nr {i} ({tsummary}) "
+                    f"Transaction result nr {i} ({txshort}) "
                     f"doesn't have a 'tsourcepos', skipping"
                 )
                 continue
@@ -180,19 +188,35 @@ def cli(cli_args=sys.argv[1:]):
                     (pos.get("sourceName", "") or "") for pos in tsourcepos
                 )
                 if len(sourceNames) != 1:
-                    raise TypeError
+                    raise TypeError(
+                        f"Not exactly one sourceName, but {sourceNames = }"
+                    )
+                sourceName = next(iter(sourceNames))
+                if not os.path.exists(sourceName):
+                    raise TypeError(
+                        f"Source file {sourceName!r} doesn't exist. "
+                        + (
+                            f"Forecasted transactions or transactions "
+                            "from STDIN can't be `hledger edit`ed."
+                            if not sourceName
+                            else ""
+                        )
+                    )
                 sourceLines = set(pos.get("sourceLine") for pos in tsourcepos)
                 if not len(sourceLines) == 2 or not all(
                     isinstance(n, int) for n in sourceLines
                 ):
-                    raise TypeError
-            except TypeError:
-                logger.info(
-                    f"Transaction result nr {i} ({tsummary}) "
-                    f"has weird {tsourcepos = }, skipping"
+                    raise TypeError(
+                        "Not exactly 2 integer source lines, "
+                        "but {sourceLines = }"
+                    )
+            except TypeError as e:
+                logger.warning(
+                    f"Transaction result nr {i} ({txshort}) "
+                    f"has weird {tsourcepos = }: {e} - Skipping"
                 )
                 continue
-            location = tuple(list(sourceNames) + sorted(sourceLines))
+            location = tuple([sourceName] + sorted(sourceLines))
             logger.debug(f"Transaction nr {i} ({txshort}) {location = }")
             yield location
 
@@ -212,7 +236,10 @@ def cli(cli_args=sys.argv[1:]):
     sourceFileContents = dict()
     for sourceFile, positions in positionsInSourceFiles.items():
         with open(sourceFile) as fh:
-            logger.info(f"Collecting transactions from file {sourceFile!r}")
+            logger.info(
+                f"Collecting {len(positions)} "
+                f"transactions from file {sourceFile!r}"
+            )
             content = fh.read()
             sourceFileContents[sourceFile] = content
             lines = splitlines(content)
@@ -272,19 +299,36 @@ def cli(cli_args=sys.argv[1:]):
         lineStartPattern = re.compile(
             r"^#\s+(?P<type>start|end)s\s+line\s+(?P<line>\d+)\s+in\s+(?P<file>.*?)\s*$"
         )
-        sourceFile, lineStart, lineEnd, newContentLines = None, None, None, []
+        sourceFile, lineStart, lineEnd, newContentLines, orphanLines = (
+            None,
+            None,
+            None,
+            [],
+            [],
+        )
 
         def print_content():
-            console.print(
-                Panel(
-                    os.linesep.join(
-                        newContentLines,
-                    ),
-                    title="Here is the content so far in case you need it",
-                    subtitle=f"{sourceFile = }, {lineStart = }, {lineEnd = }",
-                    box=rich.box.HORIZONTALS,
+            if any(newContentLines):
+                console.print(
+                    Panel(
+                        os.linesep.join(
+                            newContentLines,
+                        ),
+                        title="Here is the content so far in case you need it",
+                        subtitle=f"{sourceFile = }, {lineStart = }, {lineEnd = }",
+                        box=rich.box.HORIZONTALS,
+                    )
+                )
+            if any(orphanLines):
+                console.print(
+                    Panel(
+                        os.linesep.join(
+                            orphanLines,
+                        ),
+                        title="Here are the orphaned lines",
+                        box=rich.box.HORIZONTALS,
+                    )
                 )
-            )
 
         for linenr, line in enumerate(
             splitlines(editorContentModified), start=1
@@ -349,6 +393,8 @@ def cli(cli_args=sys.argv[1:]):
                     continue
             elif lineStart is not None:
                 newContentLines.append(line)
+            else:
+                orphanLines.append(line)
         if any((sourceFile, lineStart, lineEnd, newContentLines)):
             logger.error(f"Remainder found. Ignoring it.")
             print_content()
