@@ -18,10 +18,15 @@ import subprocess
 import logging
 import functools
 import itertools
+import shlex
 from contextlib import contextmanager
 
 # internal modules
-from hledger_utils.utils import flatten
+from hledger_utils.utils import (
+    flatten,
+    handle_keyboardinterrupt,
+    handle_exception,
+)
 
 # external modules
 import pandas as pd
@@ -32,17 +37,25 @@ import psutil
 import numpy as np
 
 import rich
+from rich import print
 from rich.console import Console
 from rich.logging import RichHandler
-from rich import print
+from rich.panel import Panel
+from rich.pretty import Pretty
+from rich.syntax import Syntax
 
 console = Console(stderr=True)
 
 logger = logging.getLogger("hledger plot")
 
 
+@contextmanager
+def nothing():
+    yield
+
+
 def str_to_str_mapping(s):
-    if m := re.fullmatch(r"^(?P<old>.*?)\s*(?:[👉⮕→➡️]|->)+\s*(?P<new>.*)$", s):
+    if m := re.fullmatch(r"^(?P<old>.+?)\s*(?:[👉⮕→➡️]|->)+\s*(?P<new>.+)$", s):
         return m.groups()
     else:
         raise argparse.ArgumentTypeError(
@@ -50,9 +63,22 @@ def str_to_str_mapping(s):
         )
 
 
+def str_times_float_to_str_mapping(s):
+    if m := re.fullmatch(
+        r"^(?P<old>.+?)\s*([*·×✖️])\s*(?P<factor>(?:\d+(?:[,.]\d*))|(?:\d*[,.]\d+))\s*(?:[👉⮕→➡️]|->)+\s*(?P<new>.+)$",
+        s,
+    ):
+        d = m.groupdict()
+        return d["old"], float(d["factor"]), d["new"]
+    else:
+        raise argparse.ArgumentTypeError(
+            f"Format: 'OLDNAME * FLOAT -> NEWNAME', not {s!r}"
+        )
+
+
 def regex_to_str_mapping(s):
     if m := re.fullmatch(
-        r"^(?P<pattern>.*?)\s*(?:[👉⮕→➡️]|->)+\s*(?P<name>.*)$", s
+        r"^(?P<pattern>.+?)\s*(?:[👉⮕→➡️]|->)+\s*(?P<name>.+)$", s
     ):
         pattern, name = m.groups()
         try:
@@ -69,7 +95,7 @@ def regex_to_str_mapping(s):
 
 def regex_to_json_dict_mapping(s):
     if m := re.fullmatch(
-        r"^(?P<pattern>.*?)\s*(?:[👉⮕→➡️]|->)+\s*(?P<json>.*)$", s
+        r"^(?P<pattern>.+?)\s*(?:[👉⮕→➡️]|->)+\s*(?P<json>.+)$", s
     ):
         pattern, jsonstr = m.groups()
         try:
@@ -103,6 +129,13 @@ def regex(s):
     return pattern
 
 
+def str_with_index_list(s):
+    return [
+        (m, handle_exception(None, int, i))
+        for m, i in re.findall(r"([a-z]+)(?:\((\d+)\))?", s)
+    ]
+
+
 parser = argparse.ArgumentParser(
     description="""
 
@@ -220,6 +253,14 @@ parser_styling_group.add_argument(
     "Can be specified multiple times.",
     default=[],
 )
+parser_styling_group.add_argument(
+    "--account-format",
+    help="Format for account name including the currency. "
+    "Default is '{account} ({commodity})' "
+    "if there are multiple currencies, otherwise just '{account}'. "
+    "Applied before any other modifications/matching.",
+    default=None,
+)
 
 
 # modification
@@ -228,7 +269,12 @@ parser_modify_group = parser.add_argument_group(
     description="Options for manipulating the data",
 )
 parser_modify_group.add_argument(
-    "--invert", help="invert amounts (done first)", action="store_true"
+    "--invert",
+    type=regex,
+    help="patterns invert amounts",
+    action="append",
+    nargs="?",
+    default=[],
 )
 parser_modify_group.add_argument(
     "--rename",
@@ -243,17 +289,72 @@ parser_modify_group.add_argument(
     metavar="REGEX -> NEWNAME",
     type=regex_to_str_mapping,
     action="append",
+    nargs="?",
     help="Mapping like 'REGEX -> NAME' to sum matching columns into a new field. "
     "Can be specified multiple times. "
-    "--sum is performed after --rename",
+    "Use e.g. --sum '.* -> total' to get a total value.",
+    default=[],
+)
+parser_modify_group.add_argument(
+    "--mean",
+    metavar="REGEX -> NEWNAME",
+    type=regex_to_str_mapping,
+    action="append",
+    nargs="?",
+    help="Mapping like 'REGEX -> NAME' to average matching columns into a new field. "
+    "Can be specified multiple times. ",
+    default=[],
+)
+parser_modify_group.add_argument(
+    "--multiply",
+    metavar="NAME * FLOAT -> NEWNAME",
+    type=str_times_float_to_str_mapping,
+    action="append",
+    help="Mapping like 'REGEX * FLOAT -> NAME' "
+    "to multiply a column with a value "
+    "and store it into a new column. "
+    "Can be specified multiple times. ",
     default=[],
 )
 parser_modify_group.add_argument(
     "--resample",
     metavar="INTERVAL",
+    action="append",
     help="DataFrame.resample() argument for data resampling "
     "(e.g. '60d' for a 60-day mean, "
     "see https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.resample.html)",
+    default=[],
+)
+parser_modify_group.add_argument(
+    "--drop",
+    metavar="REGEX",
+    help="regular expressions for dropping columns right before plotting. "
+    "Can be specified multiple times.",
+    action="append",
+    type=regex,
+    default=[],
+)
+parser_modify_group.add_argument(
+    "--only",
+    metavar="REGEX",
+    help="regular expressions for keeping only matching columns for plotting. "
+    "Can be specified multiple times.",
+    action="append",
+    type=regex,
+    default=[],
+)
+parser_modify_group.add_argument(
+    "--mod-order",
+    metavar="LIST",
+    default=str_with_index_list(
+        default := "rename invert sum multiply resample mean drop only"
+    ),
+    help="List of modification steps like 'sum(1),rename(2),resample(1)' "
+    "specifying the application order of modifiers. "
+    "The indices refer to the Nth instruction of that type (1-based). "
+    "Without an index, all modifiers of that type are executed in order. "
+    f"Default order: {default!r}",
+    type=str_with_index_list,
 )
 
 
@@ -269,6 +370,7 @@ hledger_subparsers = hledger_parser.add_subparsers(
 hledger_balance_subparser = hledger_subparsers.add_parser(
     "balance", aliases=["b", "bal"]
 )
+hledger_balance_subparser.add_argument("--layout")
 hledger_balance_subparser.add_argument(
     "--historical", "-H", action="store_true"
 )
@@ -301,6 +403,12 @@ hledger_subparsers.add_parser("tags")
 hledger_subparsers.add_parser("web")
 
 
+@handle_keyboardinterrupt(
+    console=console,
+    exitcode=handle_exception(
+        0, int, os.environ.get("HLEDGER_PLOT_KEYBOARDINTERRUPT_EXITCODE")
+    ),
+)
 def cli(cli_args=sys.argv[1:]):
     args, hledger_args = parser.parse_known_args()
     logging.basicConfig(
@@ -323,6 +431,24 @@ def cli(cli_args=sys.argv[1:]):
             logging.getLogger(name).setLevel(10000)
 
     logger.debug(f"{args = }")
+    if args.invert == [None]:
+        args.invert = [re.compile(r".*")]
+        logger.debug(
+            f"Assuming you want to --invert all columns, "
+            "setting patterns to {args.invert}"
+        )
+    if args.sum == [None]:
+        args.sum = [(re.compile(r".*"), "∑ sum")]
+        logger.debug(
+            f"Assuming you want to --sum all columns, "
+            "setting mapping to {args.sum}"
+        )
+    if args.mean == [None]:
+        args.mean = [(re.compile(r".*"), "⌀ mean")]
+        logger.debug(
+            f"Assuming you want to --mean all columns, "
+            "setting mapping to {args.mean}"
+        )
     logger.debug(f"{hledger_args = }")
 
     try:
@@ -337,6 +463,8 @@ def cli(cli_args=sys.argv[1:]):
             )
         )
         sys.exit(1)
+    logger.debug(f"{hledger_parsed_args = }")
+    logger.debug(f"{hledger_unknown_args = }")
 
     # TODO: automatically get aliases from above?
     parseable_hledger_commands = ["balance", "b", "bal"]
@@ -348,10 +476,9 @@ def cli(cli_args=sys.argv[1:]):
         )
         sys.exit(1)
 
-    if os.path.realpath(sys.argv[0]) == os.path.realpath(__file__):
-        hledger_executable = "hledger"
-    else:
-        hledger_parent_process = psutil.Process(os.getppid())
+    hledger_executable = "hledger"
+    hledger_parent_process = psutil.Process(os.getppid())
+    if "hledger" in hledger_parent_process.cmdline()[0]:
         hledger_executable = hledger_parent_process.cmdline()[0]
 
     hledger_cmdline_parts = [hledger_executable] + hledger_args
@@ -365,6 +492,13 @@ def cli(cli_args=sys.argv[1:]):
 
     hledger_cmdline_extra_args = []
     if "balance".startswith(hledger_parsed_args.command):
+        if (layout := hledger_parsed_args.layout) and layout != "tidy":
+            logger.error(
+                "Please don't specify something else than --layout=tidy"
+            )
+            sys.exit(1)
+        else:
+            hledger_cmdline_extra_args.append("--layout=tidy")
         if not hledger_parsed_args.historical:
             logger.info(
                 "ℹ️  Hint: You might want to consider adding --historical/-H to get the real balances at these times"
@@ -381,64 +515,135 @@ def cli(cli_args=sys.argv[1:]):
             logger.info("Adding --daily aggregation period for you")
             hledger_cmdline_extra_args.append("--daily")
 
-    hledger_cmdline_parts.extend(hledger_cmdline_extra_args)
-    hledger_cmdline = " ".join(hledger_cmdline_parts)
-    logger.info("🚀  Executing {}".format(repr(hledger_cmdline)))
-    try:
-        hledger = subprocess.Popen(
-            hledger_cmdline_parts,
-            stdout=subprocess.PIPE,
-            stderr=subprocess.PIPE,
+    def hledger_to_dataframe(hledger_cmdline_parts):
+        hledger_cmdline_parts.extend(hledger_cmdline_extra_args)
+        hledger_cmdline = shlex.join(hledger_cmdline_parts)
+        logger.info("🚀  Executing {}".format(repr(hledger_cmdline)))
+        try:
+            hledger = subprocess.Popen(
+                hledger_cmdline_parts,
+                stdout=subprocess.PIPE,
+                stderr=subprocess.PIPE,
+                encoding="utf-8",
+                errors="ignore",
+            )
+        except BaseException as e:
+            logger.info(
+                "Couldn't execute {}: {}".format(repr(hledger_cmdline), e)
+            )
+            sys.exit(1)
+
+        # TODO: Reading the whole output at once is ridiculous.
+        # WAY better: a file-object wrapper that sanitizes the output on-the-fly so that
+        # pandas can then directly parse it. But I couldn't get that to work AT ALL...
+
+        logger.info("📤  Reading hledger's output...")
+        hledger_output, hledger_stderr = hledger.communicate()
+
+        logger.debug(f"{hledger_stderr = }")
+        if hledger_stderr:
+            logger.warning(
+                f"Command >>> {hledger_cmdline} <<< returned STDERR:\n{hledger_stderr}"
+            )
+
+        if not hledger_output:
+            logger.info(
+                f"Command >>> {hledger_cmdline} <<< returned no output."
+            )
+            sys.exit(0)
+
+        if logger.getEffectiveLevel() < logging.DEBUG:
+            console.print(
+                Panel(
+                    Syntax(
+                        hledger_output,
+                        "python",
+                        line_numbers=True,
+                    ),
+                    title=f"Output of {hledger_cmdline}",
+                )
+            )
+
+        logger.debug("Parsing output...")
+        try:
+            read_csv_args, read_csv_kwargs = [
+                io.StringIO(hledger_output)
+            ], dict(
+                on_bad_lines="skip",
+            )
+            logger.debug(
+                f"Running pd.read_csv(*{read_csv_args}, **{read_csv_kwargs})"
+            )
+            data = pd.read_csv(*read_csv_args, **read_csv_kwargs)
+
+        except Exception as e:
+            logger.exception(f"Error: {e}")
+            logger.info(hledger_output)
+            sys.exit(1)
+
+        logger.debug("Converting times...")
+        for col in ["period", "start_date", "end_date"]:
+            if col in data:
+                data[col] = pd.to_datetime(data[col], errors="coerce")
+
+        logger.debug("Converting amounts...")
+        for col in ["value"]:
+            if col in data:
+                data[col] = pd.to_numeric(data[col], errors="coerce")
+
+        return data
+
+    data = hledger_to_dataframe(hledger_cmdline_parts)
+    commodities = list(data.commodity.unique())
+    if not args.account_format:
+        args.account_format = (
+            "{account} ({commodity})" if len(commodities) > 1 else "{account}"
         )
-    except BaseException as e:
-        logger.info("Couldn't execute {}: {}".format(repr(hledger_cmdline), e))
-        sys.exit(1)
 
-    # TODO: Reading the whole output at once is ridiculous.
-    # WAY better: a file-object wrapper that sanitizes the output on-the-fly so that
-    # pandas can then directly parse it. But I couldn't get that to work AT ALL...
+    # Try to auto-format commodities right
+    # TODO: Can be prevent this loop by using `hledger stats` for example to
+    # know the commodities in advance? Although hledger stats also parses
+    # everything so might not be much of a benefit...
+    if missing_values := data["value"].isna().sum():
+        logger.warning(
+            f"There are {missing_values} NaNs in the converted amounts."
+        )
+        commodity_format_opts = list(
+            itertools.chain.from_iterable(
+                ("-c", f'1000000.0000 "{curr}"') for curr in commodities
+            )
+        )
+        logger.info(f"Retrying with {shlex.join(commodity_format_opts)}")
+        data = hledger_to_dataframe(
+            hledger_cmdline_parts + commodity_format_opts
+        )
+        if missing_values := data["value"].isna().sum():
+            logger.warning(
+                f"Didn't help, trying to plot anyway but it'll look weird."
+            )
+        else:
+            logger.info(f"That worked, no NaNs in the amounts anymore! 🎉 ")
 
-    logger.info("📤  Reading hledger's output...")
-    hledger_output, hledger_stderr = hledger.communicate()
-    hledger_output = hledger_output.decode()
+    time_columns = [col for col in data if hasattr(data[col], "dt")]
+    logger.debug(f"Using minimum of {time_columns = } as time")
+    data = data.set_index(data[time_columns].min(axis="columns"))
+    data = data.drop(columns=time_columns)
+    logger.debug(f"data = \n{data}")
 
-    logger.debug("Sanitizing output...")
-    # floating-point numbers (just something like " - 1000,00 €" is possible)
-    hledger_output = re.sub(
-        r'"(\s*-?\s*[.\d]+)[,.](\d+)\s*[^"]*"', r"\1.\2", hledger_output
+    data["account"] = data["account"].astype(str)
+
+    groupby = ["account", "commodity"]
+    logger.info(f"Grouping accounts and commodities by {groupby}")
+    data = pd.concat(
+        [
+            g["value"].rename(
+                args.account_format.format(**dict(zip(groupby, n)))
+            )
+            for n, g in data.groupby(groupby)
+        ],
+        axis="columns",
     )
-    # integer numbers
-    hledger_output = re.sub(
-        r'"(\s*-?\s*\d+)\s*[^"\d]*"', r"\1", hledger_output
-    )
-    # dates
-    hledger_output = re.sub(
-        r"(\d{4}\D\d{2}\D\d{2})[a-z]\d+",
-        r"\1",
-        hledger_output,
-        flags=re.IGNORECASE,
-    )
-
-    logger.debug("Parsing output...")
-    try:
-        data = pd.read_csv(
-            io.StringIO(hledger_output), index_col=0, on_bad_lines="skip"
-        ).T
-    except BaseException as e:
-        logger.exception(f"Error: {e}")
-        logger.info(hledger_output)
-        sys.exit(1)
-
-    logger.debug("Converting times...")
-    try:
-        data.index = pd.to_datetime(data.index)
-    except BaseExeption as e:
-        logger.info("Error: {}".format(repr(e)))
-        logger.info(data)
-        sys.exit(1)
-
-    for col in data:
-        data[col] = pd.to_numeric(data[col], errors="coerce")
+    logger.debug(f"data = \n{data}")
 
     # some sane defaults for rcParams
     plt.rcParams["legend.handlelength"] = 5
@@ -470,23 +675,26 @@ def cli(cli_args=sys.argv[1:]):
     plt.rcParams.update(args.rcParams)
 
     # 🔢 Data Modification
-    if args.invert:
-        logger.info(f"↔️  Inverting amounts")
-        data = data * -1
+    def invert_handler(patterns, data=data):
+        for pattern in patterns:
+            for col in data:
+                if m := pattern.search(col):
+                    logger.info(f"↔️  Inverting {col!r} amount")
+                    data[col] = data[col] * -1
+        return data
 
-    logger.debug(f"{data = }")
+    # rename nan to actual "nan"
     data = data.rename(columns={x: str(x) for x in data})
-    if args.rename:
-        logger.debug(f"{data.columns = }")
-        renames = dict(args.rename)
-        logger.info(f"Renaming columns: {renames}")
-        data = data.rename(columns=renames)
-        logger.debug(f"{data = }")
 
-    # TODO: args.sum
-    if args.sum:
-        logger.debug(f"{args.sum = }")
-        for pattern, newname in args.sum:
+    def rename_handler(mapping, data=data):
+        logger.debug(f"{mapping = }")
+        renames = dict(mapping)
+        logger.info(f"Renaming columns: {renames}")
+        return data.rename(columns=renames)
+
+    def sum_handler(mapping, data=data):
+        logger.debug(f"Applying --sum to {mapping = }")
+        for pattern, newname in mapping:
             if columns := list(filter(pattern.search, data)):
                 data[newname] = np.nansum(
                     [data[col] for col in columns], axis=0
@@ -498,13 +706,101 @@ def cli(cli_args=sys.argv[1:]):
                 logger.info(
                     f"🤷  No columns matching pattern {pattern.pattern!r}"
                 )
+        return data
 
-    if args.resample:
-        data = data.resample(args.resample).sum()
+    def mean_handler(mapping, data=data):
+        logger.debug(f"{mapping = }")
+        for pattern, newname in mapping:
+            if columns := list(filter(pattern.search, data)):
+                data[newname] = np.nanmean(
+                    [data[col] for col in columns], axis=0
+                )
+                logger.info(
+                    f"Created new column {newname!r} averaging {columns}"
+                )
+            else:
+                logger.info(
+                    f"🤷  No columns matching pattern {pattern.pattern!r}"
+                )
+        return data
 
-    @contextmanager
-    def nothing():
-        yield
+    def multiply_handler(mapping, data=data):
+        logger.debug(f"{mapping = }")
+        for oldname, factor, newname in mapping:
+            if oldname not in data:
+                logger.error(
+                    f"No such column {oldname!r} to multiply with {factor}"
+                )
+                continue
+            data[newname] = data[oldname] * factor
+        return data
+
+    def resample_handler(intervals, data=data):
+        logger.debug(f"{intervals = }")
+        for interval in intervals:
+            data = data.resample(interval).sum()
+        return data
+
+    def drop_handler(mapping, data=data):
+        for pattern in mapping:
+            if columns := list(filter(pattern.search, data)):
+                data = data.drop(columns=columns)
+                logger.info(f"Dropped columns {columns!r}")
+            else:
+                logger.info(
+                    f"🤷  No columns matching pattern {pattern.pattern!r}"
+                )
+        return data
+
+    def only_handler(patterns, data=data):
+        if columns := list(
+            col for col in data if any(p.search(col) for p in patterns)
+        ):
+            data = data[columns]
+            logger.info(f"Kept only columns {columns!r}")
+        else:
+            logger.info(f"🤷  No columns matching {patterns = }")
+        return data
+
+    modhandlers = dict(
+        sum=sum_handler,
+        invert=invert_handler,
+        rename=rename_handler,
+        mean=mean_handler,
+        multiply=multiply_handler,
+        drop=drop_handler,
+        only=only_handler,
+        resample=resample_handler,
+    )
+
+    for step, (modname, index) in enumerate(args.mod_order, start=1):
+        logger.debug(f"--mod-order step Nr. {step}: {modname = !r} {index = }")
+        if (handler := modhandlers.get(modname)) and (
+            ((allmodsteps := getattr(args, modname, None)) is not None)
+        ):
+            logger.debug(
+                f"All {len(allmodsteps)} steps for {modname!r}: {allmodsteps = }"
+            )
+            if (
+                1
+                <= (index if isinstance(index, int) else -1)
+                <= len(allmodsteps)
+            ):
+                modsteps = allmodsteps[(index - 1) : index]
+            else:
+                modsteps = allmodsteps
+            if modsteps:
+                logger.debug(f"Running {modname!r} steps {modsteps}")
+                data = handler(modsteps, data=data)
+                logger.debug(
+                    f"data after {modname!r} with steps {modsteps}:\n{data}"
+                )
+            else:
+                logger.warning(f"No {modname!r} steps? ({index = })")
+        else:
+            logger.warning(
+                f"No handler for --mod-order step {modname = !r} {index = } ({allmodsteps = })"
+            )
 
     logger.info("📈  Plotting...")
     with plt.xkcd() if args.xkcd else nothing():
@@ -575,6 +871,7 @@ def cli(cli_args=sys.argv[1:]):
         if args.title:
             fig.suptitle(args.title)
         ax.set_title(args.axtitle or " ".join(hledger_args))
+        ax.set_ylabel(" ".join(f"[{c}]" for c in commodities))
         ax.legend(ncols=math.ceil(len(ax.get_lines()) / 40))
 
         fig.tight_layout()
